import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import _thread
from threading import Event
from time import sleep, strftime
from copreus.drivers.adc import ADC
from tests_manual.tools import resister_to_pubs, send_to_topic, get_topics_sub, get_device_config, get_topics_pub

filename = 'config_adc.yaml'

send_to_topic(filename, "/test/closed", 1)

# check if use-calibration is set to 'false' ----------
c = get_device_config(filename)
if c["use-calibration"]:
    raise ValueError("for calibration parameter 'use-calibration' must be set to false "
                     "(currently: {}).".format(c["use-calibration"]))
# -----------------------------------------------------


# on_message ------------------------------------------
volt_in = -1.0
received = Event()
received.clear()
topic_volt = get_topics_pub(filename)["volt"]

def volt_in_received(client, userdata, msg):
    if msg.topic == topic_volt:
        global volt_in
        volt_in = float(msg.payload)
        received.set()


resister_to_pubs(filename,on_message=volt_in_received)
# -----------------------------------------------------


# create csv file -------------------------------------
data_adc = open("calibration_adc_"+strftime("%Y%m%d-%H%M%S")+".csv", "w")
# -----------------------------------------------------


# start ADC driver ------------------------------------
_adc_thread = _thread.start_new_thread(ADC.standalone, (['-c', filename],))
topic_readnow = get_topics_sub(filename)["readnow"]
# -----------------------------------------------------

# get number of steps as user input -------------------
sleep(1)  # wait a little bit - previous tasks should be finished before continuing
steps = -1
try:
    while steps <= 0:
        newstep = input("Number of steps (0<x<=64): ")
        if int(newstep) > 0 and int(newstep) <= 64:
            steps = int(newstep)
            print("new value for step: {}.".format(steps))
except ValueError:
    pass
# -----------------------------------------------------

print("start calibration -------------------------------")
print("volt_out;volt_in;multimeter")
print("# - [ref_value, raw_value]", file=data_adc)

for step in range(1, steps+1):
    input("to start next iteration ({}/{}) press enter.".format(step, steps))

    # read value from adc ---------------------------------
    received.clear()
    send_to_topic(filename, topic_readnow, 1)
    received.wait()
    # -----------------------------------------------------

    # get reference value from user -----------------------
    multimeter = -1
    while multimeter < 0:
        mm = input("multimeter value for adc value {} V: ".format(volt_in))
        try:
            mm = float(mm)
            accept_value = input("accept entered value {} V (y/N)? ".format(mm))
            if accept_value == "y" or accept_value == "Y":
                multimeter = mm
        except ValueError:
            pass
    # -----------------------------------------------------

    # write to file/stdout --------------------------------
    print("{};{};{}".format(step, volt_in, multimeter))
    # - [ref_value, raw_value]
    print(" - [{}, {}]".format(multimeter, volt_in), file=data_adc)
    # -----------------------------------------------------

data_adc.close()
print("done. -------------------------------------------")
print("cleaning up.")
try:
    _thread.interrupt_main()
except KeyboardInterrupt:
    pass
sleep(1)
print("finished.")









from time import time
import RPi.GPIO as GPIO
from copreus.baseclasses.adriver import ADriver
from copreus.baseclasses.aevents import AEvents
from copreus.schema.rotaryencoder import get_schema


class RotaryEncoder(ADriver, AEvents):
    """Driver for rotary encoder like the KY-040 with software solutions for debouncing and direction detection.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_pub:
          * rotate - mqtt-translations.rotate-left and mqtt-translations.rotate-right
          * button_pressed - mqtt-translations.button_pressed
          * button_state - mqtt-translations.button_state-open and mqtt-translations.button_state-closed
      * RotaryEncoder entries
        * sw: gpio pin for pressing the rotary encoder
        * dt: dt gpio pin
        * clk: clk gpio pin

    Example:
    device:
        type: rotaryencoder
        pin_sw:  12
        pin_dt:  16
        pin_clk: 20
        topics-pub:
            rotate: /input1/rotate
            button_pressed: /input1/button/pressed
            button_state:   /input1/button/state
        mqtt-translations:
            rotate-left: LEFT
            rotate-right: RIGHT
            button_pressed: PRESSED
            button_state-open: OPEN
            button_state-closed: CLOSED

    """

    _clk = -1  # clk gpio pin id
    _dt = -1  # dt gpio pin id
    _sw = -1  # sw gpio pin id
    _last_clk = -1  # last clk value - used for rotation direction detection
    _last_time = -1 # time stamp of the last rotation step
    _last_direction = 0  # rotation direction of the last rotation step
    _min_time_gap = 0.100  # minimum time gap between direction changes
    _LEFT = -1  # internal constant - rotation direction left
    _RIGHT = 1  # internal constant - rotation direction right

    def __init__(self, config, mqtt_client=None, logger=None):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=__name__)
        AEvents.__init__(self, self._config, self._logger)

        self._clk = self._config["pin_clk"]
        self._dt = self._config["pin_dt"]
        self._sw = self._config["pin_sw"]

        self._add_event(self._clk, self._callback_rotary, 1)
        self._add_event(self._sw, self._callback_sw)

        self._last_time = time()

    def _callback_sw(self, channel):
        """Event handler for gpio pin 'sw'. Publishes to the topics 'button_state' and 'button_pressed'."""
        self._logger.info("SW-event detected.")
        state = GPIO.input(self._sw)
        if not state:
            self._publish_value(self._topics_pub["button_pressed"], self._mqtt_translations["button_pressed"])
            self._publish_value(self._topics_pub["button_state"], self._mqtt_translations["button_state-closed"])
        else:
            self._publish_value(self._topics_pub["button_state"], self._mqtt_translations["button_state-open"])

    def _callback_rotary(self, channel):
        """Event handler for gpio pin 'clk'. Detects the rotation direction and filters out 'rotation bounces' -
        rotation direction changes that happen to fast und are thus considered to originiate in bounces. Publishes
        to the topic 'rotate'."""
        akt_clk = GPIO.input(self._clk)
        akt_time = time()
        self._logger.info("Rotary-event detected.")

        if akt_clk != self._last_clk:
            rotate = self._LEFT
            if GPIO.input(self._dt) != akt_clk:
                rotate = self._RIGHT

            #filter
            if akt_time - self._last_time < self._min_time_gap and rotate != self._last_direction:
                self._logger.info("Time between direction change to small - skipping this event.")
            else:
                if rotate == self._RIGHT:
                    self._publish_value(self._topics_pub["rotate"], self._mqtt_translations["rotate-right"])
                else:
                    self._publish_value(self._topics_pub["rotate"], self._mqtt_translations["rotate-left"])
                self._last_direction = rotate
                self._last_time = akt_time
        self._last_clk = akt_clk

    def _driver_start(self):
        """ADriver._driver_start"""
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(self._dt, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.setup(self._sw, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        GPIO.setup(self._clk, GPIO.IN, pull_up_down=GPIO.PUD_UP)
        self._last_clk = GPIO.input(self._clk)
        self._register_events()

    def _driver_stop(self):
        """ADriver._driver_stop"""
        self._unregister_events()
        GPIO.cleanup(self._dt)
        GPIO.cleanup(self._sw)
        GPIO.cleanup(self._clk)

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method RotaryEncoder.standalone()."""
    RotaryEncoder.standalone()


if __name__ == "__main__":
    RotaryEncoder.standalone()



def get_schema_parts(propertyname):
    return {
        "properties": {
            propertyname: {
                "description": "calibration",
                "type": "object",
                "properties": {
                    "values": {
                        "description": "list of calibration values",
                        "type": ["array", "null"],
                        "items": {
                            "type": "array",
                            "items": [
                                {"type": "number"},
                                {"type": "number"}
                            ]
                        }
                    },
                    "use-calibration": {
                        "description": "if set to False, calibration will be ommitted.",
                        "type": "boolean"
                    }
                },
                "required": ["values", "use-calibration"],
                "additionalProperties": False
            },
        },
        "required": [propertyname],
    }


from copreus.baseclasses.aepaper import EPaperMQTTMessageConverter
from copreus.baseclasses.aepaper import AEPaper
from PIL import Image
from io import BytesIO
import threading
import datetime
import queue
from copreus.schema.epapersimple import get_schema


class EPaperSimple(AEPaper):
    """
    A simplified interface for the epaper. It accepts full images only and displays the last recevied message. If
    during an update several other images are received all but the last one are dropped. Additionally, the display
    can be wiped regularly to ensure that no fragments/ghosts are shown.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_sub:
            * image - a single image covering the whole display to be placed in the current buffer.
        * topics_pub:
            * message_queue_size - publishes the number of messages that wait to be processes.
      * EPaperSimple entries
        * wipe-screen:
            * every-nth-day - how often should the timer be called
            * time - time during a day when the wipe screen action should take place. best use a time when no one will
            look on the screen like early in the morning
            * at-start-up - wipe the screen immedietly after start of the driver

    Example:
        device:
            type: epapersimple
            model: 2.9
            spi:
                pin_cs: -1 # use spi cs mechanism. GPIO08/SPI_CE0_N
                bus: 0
                device: 0
                maxspeed: 2000000
            transpose: 270
            pin_rst:  13
            pin_dc:   19
            pin_busy: 6
            VCOM: -3.2
            autodeepsleep: True
            topics-sub:
                image: /test/display/full_image
            topics-pub:
                message_queue_size: /test/display/message_queue_size
            wipe-screen:
                every-nth-day: 1  # 0 for never
                time: 03:15 # wipe-screen will be called at the first update after this time. ignored if every-nth-day==0
                at-start-up: True

    """
    _topic_pub_msg_queue_size = None  # number of images in queue - possible values are 0 or 1
    _akt_image_entry = None  # last displayed image

    _update_available = None  # event - triggers if a new image is available -> display image is called by worker
    _image_queue = None

    _display_lock = None  # lock to restrict access to the display - used by wipe-screen and display image
    _stop_threads = None  # threading.Event to signal the poll loop to stop immediately.
    _display_update_thread = None  # thread in which the poll loop is executed.
    _wipe_screen_on_start = None  # wipe screen immediately
    _wipe_screen_thread = None  # thread for timed wipe-screen
    _wipe_screen_time = None  # time HH:MM at which the screen should be wiped
    _wipe_img_black = None  # black image for wiping
    _wipe_img_white = None  # white image for wiping

    def __init__(self, config, mqtt_client=None, logger=None, spi_lock=None):
        """
        Constructor

        :param config: yaml config structure
        :param mqtt_client: mymqttclient instance (optional)
        :param logger: logger instance (optional)
        :param spi_lock: spi lock instance (optional)
        """

        AEPaper.__init__(self, config, mqtt_client, logger, spi_lock)

        self._update_available = threading.Event()
        self._update_available.clear()
        self._display_lock = threading.Lock()
        self._stop_threads = threading.Event()
        self._display_update_thread = threading.Thread(target=self._display_akt_image)
        self._image_queue = queue.Queue()

        self._wipe_screen_on_start = bool(self._config["wipe-screen"]["at-start-up"])
        self._every_nth_day = self._config["wipe-screen"]["every-nth-day"]
        temp = self._config["wipe-screen"]["time"].split(":")
        self._wipe_screen_time = datetime.time(hour=int(temp[0]), minute=int(temp[1]))
        self._wipe_screen_thread = threading.Thread(target=self._wipe_screen_timer)

        self._wipe_img_black = Image.new('L', (self._width, self._height), 0)
        self._wipe_img_white = Image.new('L', (self._width, self._height), 255)

        self._topic_pub_msg_queue_size = self._topics_pub["message_queue_size"]

    def _set_update_available(self):
        """sets state to none emtpy queue (sets _update_available event and publishes message queue size 1."""
        self._logger.info("EPaperSimple._set_update_available")
        self._update_available.set()
        self._mqtt_client.publish(self._topic_pub_msg_queue_size, 1)

    def _clear_update_available(self):
        """sets state to empty queue (clears _update_available event and publishes message queue size 0."""
        self._logger.info("EPaperSimple._clear_update_available")
        if self._image_queue.empty():
            self._update_available.clear()
            self._mqtt_client.publish(self._topic_pub_msg_queue_size, 0)
        else:
            self._logger.info("EPaperSimple._clear_update_available - skipped. image queue not empty: {}"
                               .format(self._image_queue.qsize()))

    def _handler_display_image(self, msg):
        """on_message handler for topic sub 'full_image'"""
        self._logger.info("EPaperSimple._handler_display_image - received full_image in topic {}."
                          .format(self._topics_sub["image"]))
        if self._update_available.is_set():
            self._logger.info("EPaperSimple._handler_display_image - replacing image that waits for being displayed "
                              "already.")
        image = Image.open(BytesIO(msg))
        image, x, y = EPaperMQTTMessageConverter.transpose_image(self._transpose, image, self._width, self._height)
        self._image_queue.put([image, x, y])
        self._set_update_available()

    def _display_akt_image(self):
        """
        send received image to the epaper. uses the image that has been assigned to _akt_image. skips update if the
        value is None. loop stops if _Stop_threads is set.

        acquire the lock (_display_frame_lock), wake up from deep sleep (optional), put the image into the
        current frame buffer, display frame, send back to deep sleep (optional), wait the _update_time and then
        release the lock."""

        self._logger.info("EPaperSimple._display_akt_image - thread started")

        while not self._stop_threads.is_set():
            self._update_available.wait(timeout=0.25)  # busy waiting ... currently no elegant way to wait for two events ...
            if not self._update_available.is_set():
                continue  # timeout reached

            while not self._image_queue.empty():
                self._logger.info("... fetching entry from image queue.")
                self._akt_image_entry = self._image_queue.get()

            if self._akt_image_entry is None:
                self._logger.info("... no image received until now - nothing to be displayed ...")
                self._clear_update_available()
                continue

            image, x, y = self._akt_image_entry

            self._logger.info("display image. acquiring lock ...")
            with self._display_lock:
                self._logger.info("... lock acquired")
                if self._auto_deep_sleep:
                    self._reset()
                self._set_frame_memory(image, x, y)
                self._display_frame()
                if self._auto_deep_sleep:
                    self._deep_sleep()
                self._stop_threads.wait(timeout=self._update_time)  # wait for image to be processed or for stop_loop event
                self._logger.info("... image displayed. lock released.")

            self._clear_update_available()

        self._logger.info("EPaperSimple._display_akt_image - thread stopped")

    def _seconds_to_next_timer_event(self):
        """
        Calculates how many seconds are from now to next "HH:MM" time.

        :return: float / seconds
        """
        now = datetime.datetime.now()
        time = now.time()
        day = now.date()
        self._logger.info("EPaperSimple._seconds_to_next_timer_event - now: {}, day: {}, time: {}"
                           .format(now, day, time))
        self._logger.info("EPaperSimple._seconds_to_next_timer_event - target_time: {}, n-th day: {}"
                           .format(self._wipe_screen_time, self._every_nth_day))

        if time > self._wipe_screen_time:
            # wait until next day (or nth day) for wipe_screen_time to occur
            self._logger.info("EPaperSimple._seconds_to_next_timer_event - time > wipe_screen_time")
            target = datetime.datetime.combine(day, self._wipe_screen_time)
            target = target + datetime.timedelta(days=self._every_nth_day)
        elif time < self._wipe_screen_time:
            # wait until wipe_screen_time today
            self._logger.info("EPaperSimple._seconds_to_next_timer_event - time < wipe_screen_time")
            target = datetime.datetime.combine(day, self._wipe_screen_time)
        else:
            # its now!
            self._logger.info("EPaperSimple._seconds_to_next_timer_event - time == wipe_screen_time")
            target = now

        wait_time = target - now
        self._logger.info("EPaperSimple._seconds_to_next_timer_event - wait_time: {}.".format(wait_time))
        return wait_time.total_seconds()

    def _wipe_screen_timer(self):
        """endless loop that waits until next wipe screen time and initialize wipe screen and restarts wait.
        stops if _stop_threads is set."""
        self._logger.info("EPaperSimple._wipe_screen_timer - thread started")
        while not self._stop_threads.is_set():
            wait_time = self._seconds_to_next_timer_event()
            self._logger.warning("EPaperSimple._wipe_screen_timer - waiting {} seconds (time: {}, nth-day: {})"
                                 .format(wait_time, self._wipe_screen_time, self._every_nth_day))
            self._stop_threads.wait(timeout=wait_time)
            if not self._stop_threads.is_set():
                self._wipe_screen()
        self._logger.info("EPaperSimple._wipe_screen_timer - thread stopped")

    def _wipe_screen(self):
        """wipes the screen by displaying alternating black and white images. restores to last displayed image
        at the end by setting _update_available timer."""
        self._logger.info("EPaperSimple._wipe_screen - started")

        self._logger.info("EPaperSimple._wipe_screen - acquiring lock ...")
        with self._display_lock:
            self._logger.info("EPaperSimple._wipe_screen - ... lock acquired")
            if self._auto_deep_sleep:
                self._reset()

            # set white
            self._logger.info("EPaperSimple._wipe_screen - display white")
            self._set_frame_memory(self._wipe_img_white, 0, 0)
            self._display_frame()
            self._stop_threads.wait(timeout=self._update_time)  # wait for image to be processed or for stop_loop event

            # set black
            self._logger.info("EPaperSimple._wipe_screen - display black")
            self._set_frame_memory(self._wipe_img_black, 0, 0)
            self._display_frame()
            self._stop_threads.wait(timeout=self._update_time)  # wait for image to be processed or for stop_loop event

            for i in range(4):
                # set white
                self._logger.info("EPaperSimple._wipe_screen - display white")
                self._display_frame()
                self._stop_threads.wait(timeout=self._update_time)  # wait for image to be processed or for stop_loop event
                # set black
                self._logger.info("EPaperSimple._wipe_screen - display black")
                self._display_frame()
                self._stop_threads.wait(timeout=self._update_time)  # wait for image to be processed or for stop_loop event

            if self._auto_deep_sleep:
                self._deep_sleep()

            self._logger.info("EPaperSimple._wipe_screen - ... lock released")

        # initiate display of akt image
        self._logger.info("EPaperSimple._wipe_screen - initiate update display")
        self._set_update_available()
        self._logger.info("EPaperSimple._wipe_screen - finished")

    def _epaper_start(self):
        """starts by subscribing to image topic, optionally wipeing screen and setting timer to next wipe time."""
        self._mqtt_client.subscribe(self._topics_sub["image"], self._handler_display_image)
        self._stop_threads.clear()
        if self._wipe_screen_on_start:
            self._wipe_screen()
        if self._every_nth_day > 0:
            self._wipe_screen_thread.start()
        self._display_update_thread.start()

    def _epaper_stop(self):
        """stops by unsubscribing, and stopping threads."""
        self._mqtt_client.unsubscribe(self._topics_sub["image"], self._handler_display_image)
        self._stop_threads.set()
        self._display_update_thread.join()
        if self._every_nth_day > 0:
            self._wipe_screen_thread.join()

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method EPaperSimple.standalone()."""
    EPaperSimple.standalone()


if __name__ == "__main__":
    EPaperSimple.standalone()



from time import time
import threading


class APolling(object):
    """Additional base class for driver that need to poll in regular intervals.

    Expects 'poll-interval' in the device yaml configuration - device is polled every n seconds. Further,
    for spontaneus polling, apolling subscribes to poll-now and upon receiving of the predefined command in
    mqtt-translations, polling is started. if 'poll-interval' is set to '0' the device reacts only to poll_now mqtt
    commands.
    """

    _poll_interval = -1  # poll time in seconds
    _stop_loop = None  # threading.Event to signal the poll loop to stop immediately.
    _loop_thread = None  # thread in which the poll loop is executed.
    _logger_apolling = False  # print debugging information if set to yes.
    _mqtt_client_apolling = False  # mqtt client instance
    _poll_lock = None  # lock polling
    _poll_now_topic = None  # listen to this topic for poll now commands
    _poll_now_command = None  # if this command is received, start polling immediately

    def __init__(self, config, mqtt_client, logger):
        self._logger_apolling = logger
        self._mqtt_client_apolling = mqtt_client
        self._poll_interval = config["poll-interval"]
        self._poll_now_command = config["mqtt-translations"]["poll-now"]
        self._poll_now_topic = config["topics-sub"]["poll-now"]
        self._stop_loop = threading.Event()
        self._poll_lock = threading.Lock()

    def poll_now(self):
        with self._poll_lock:
            self._poll_device()

    def _poll_now(self, msg):
        """on_message handler for topic sub 'readnow'."""
        msg = msg.decode("utf-8")
        self._logger_apolling.info("received message '{}' on topic '{}'.".format(msg, self._poll_now_topic))
        if str(msg) == str(self._poll_now_command):
            self.poll_now()
        else:
            self._logger_apolling.error("msg expects {}. received '{}' instead.".
                               format(self._poll_now_command, msg))
            raise ValueError("msg expects {}. received '{}' instead.".
                             format(self._poll_now_command, msg))

    def _poll_loop(self):
        """Executes _poll_device and then waits (_POLL_INTERVAL - execution time of _poll_device). This guarantees that
        e.g. the state of a pin is checked exactly each 30 seconds."""
        self._logger_apolling.info("APolling._poll_loop - entered poll_loop method.")

        while not self._stop_loop.isSet():
            start = time()
            with self._poll_lock:
                self._poll_device()
            sleep_for = max(0, self._poll_interval - (time() - start))

            self._logger_apolling.info("APolling._poll_loop - sleep for " + str(sleep_for) + " seconds.")
            self._stop_loop.wait(sleep_for)

        self._logger_apolling.info("APolling._poll_loop - exiting poll_loop method.")

    # @abstract
    def _poll_device(self):
        """This method must be implemented by the silbing class. Whatever should must be done to poll the device should
        be placed here."""
        self._logger_apolling.error("APolling._poll_device - Please implement this method")
        raise NotImplementedError("Please implement this method!")

    def _start_polling(self):
        """Start a new thread with _pool_loop. Usually called in the _start_sequence method of the silbling."""
        if self._poll_interval == 0:
            self._logger_apolling.info("APolling._start_polling - poll interval is set to '0' -> poll loop is disabled.")
        else:
            self._logger_apolling.info("APolling._start_polling - start loop thread.")
            self._stop_loop.clear()
            self._loop_thread = threading.Thread(target=self._poll_loop)
            self._loop_thread.start()
        self._logger_apolling.info("APolling._start_polling - registering poll_now topic handler.")
        self._mqtt_client_apolling.subscribe(self._poll_now_topic, self._poll_now)
        self._logger_apolling.info("APolling._start_polling - started.")

    def _stop_polling(self):
        """Stop _pool_loop with Event _stop_loop. Usually called in the _stop_sequence method of the silbling."""
        self._logger_apolling.info("APolling._stop_polling - unregistering poll_now topic handler.")
        self._mqtt_client_apolling.unsubscribe(self._poll_now_topic, self._poll_now)
        if self._poll_interval > 0:
            self._logger_apolling.info("APolling._start_polling - stopping loop thread.")
            self._stop_loop.set()
            self._loop_thread.join()
        self._logger_apolling.info("APolling._stop_polling - stopped.")




import pelops.mylogger as mylogger
from enum import Enum


class _CalibrationMethod(Enum):
    NO = 0,
    OFFSET = 1,
    TWO_POINTS = 2,


class CalibratedValue(object):
    """Utility class that takes a value returns a calibrated value.

    Two calibration methods are implemented:
      0. the list calibration_data is empty - the value is returned unchanged.
      1. the list contains exactly one pair - the value is applied with the resulting offset.
      2. the list contains several pairs - the two neighboring (left and right) pairs are selected and the value
      accordingly changed and returned. (this needs an ordered list)

    Additionally a conversion factor can be provided. For example, when using an 12bit DAC for a voltage range of 0-24V
    the conversion factor is 2**12 steps / 24 V = 170.6 steps/V.

    There are two public methods:
      * value(raw) - use it to calibrate measurements. E.g. ADC
      * raw(value) - use it to convert a set point to the value for the device. E.g. DAC
    """
    _calibration_data = None  # ordered list of tuples [[ref_value, raw_value], ...]
    _conversion_factor = -1  # conversion factor
    _method = 0  # internal flag [0,1,2] - is set in __init__ and stores which calibration method should be used
    _offset_value = -1  # internal variable - stores pre-calculated offset for method 1

    _col_ref = 0  # position of reference value in tuple
    _col_raw = 1  # position of raw value in tuple

    _logger = None
    _config = None

    def __init__(self, logger, config, conversion_factor=1):
        self._config = config
        self._logger = mylogger.get_child(logger, __name__)
        self._logger.info("constructor")
        self._logger.debug("calibration_data: ".format(self._config["values"]))
        self._logger.debug("conversion_factor: ".format(conversion_factor))

        self._conversion_factor = conversion_factor
        self._calibration_data = self._config["values"]  # [[ref_value, raw_value], ...]

        if self._calibration_data is None or len(self._calibration_data) <= 0:
            self._method = _CalibrationMethod.NO
        elif len(self._calibration_data) == 1:
            self._method = _CalibrationMethod.OFFSET
            self._offset_value = self._calibration_data[0][self._col_raw] - self._calibration_data[0][self._col_ref]
        else:
            self._method = _CalibrationMethod.TWO_POINTS

        if not self._config["use-calibration"]:
            self._method = _CalibrationMethod.NO

        self._logger.info("method is set to: ".format(self._method))

    def _corrected(self, value, flip=False):
        """Returns the corrected value. Invokes the correction method according to _method.

        flip is used to change correction direction:
          * False: raw->value
          * True: value->raw"""
        if self._method == _CalibrationMethod.NO:
            value = self._no_calib(value)
        elif self._method == _CalibrationMethod.OFFSET:
            value = self._offset(value, flip)
        elif self._method == _CalibrationMethod.TWO_POINTS:
            value = self._two_points(value, flip)
        else:
            self._logger.error("don't know what to do with method id = {}.".format(self._method))
            raise ValueError("don't know what to do with method id = {}.".format(self._method))
        return value

    def _no_calib(self, value):
        """returns the unchanged value"""
        return value

    def _offset(self, value, flip):
        """Add/substract _offset_value"""
        if flip:
            value -= self._offset_value
        else:
            value += self._offset_value
        return value

    def _two_points(self, value, flip):
        """Select the surrounding calibration data points and interpolate the calibration value."""
        if flip:
            cref = self._col_raw
            craw = self._col_ref
        else:
            craw = self._col_raw
            cref = self._col_ref

        pos = 1
        while pos < len(self._calibration_data)-1:
            if value <= self._calibration_data[pos][craw]:
                break
            pos = pos + 1

        reflow = self._calibration_data[pos - 1][cref]
        rawlow = self._calibration_data[pos - 1][craw]
        refhigh = self._calibration_data[pos][cref]
        rawhigh = self._calibration_data[pos][craw]

        refrange = refhigh - reflow
        rawrange = rawhigh - rawlow
        corrected = (((value - rawlow) * refrange) / rawrange) + reflow
        return corrected

    def value(self, raw):
        """Convert raw measurements into calibrated values. If use_calibration is set to False the unchanged raw value
        is returned (for calibration/testing)."""
        value = raw * self._conversion_factor
        value = self._corrected(value, False)
        return value

    def raw(self, value):
        """Convert raw measurements into calibrated values. If use_calibration is set to False the unchanged raw value
        is returned (for calibration/testing)."""
        value = self._corrected(value, True)
        raw = value / self._conversion_factor
        return int(raw)



import RPi.GPIO as GPIO
import Adafruit_DHT
from copreus.baseclasses.adriver import ADriver
from copreus.baseclasses.apolling import APolling
from copreus.baseclasses.calibratedvalue import CalibratedValue
from copreus.schema.dht import get_schema
from copreus.baseclasses.aevents import AEvents


class DHT(ADriver, APolling, AEvents):
    """Driver for the DHT temperature/humidity sensor family.

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_pub: temperature, humidity
      * APolling entries
      * CalibratedValue entries in a sub-block named 'calibration_temperature'
      * CalibratedValue entries in a sub-block named 'calibration_humidity'
      * DHT entries
        * pin: gpio @ raspberry
        * sensor-type: DHT11, DHT22, AM2302

    Example:
    device:
        type: dht
        sensor-type: DHT22
        poll-interval: 30
        pin: 26
        topics-pub:
            temperature: /dht22/temperature/raw
            humidity: /dht22/humidity/raw
        topics-sub:
            poll-now: /dht22/pollnow
        mqtt-translations:
            poll-now: True
        use-calibration-temperature: True
        calibration-temperature:
        # - [ref_value, raw_value]
        use-calibration-humidity: True
        calibration-humidity:
        # - [ref_value, raw_value]
        event-pin:  # trigger for poll_now (optional)
            pin: 21
            flank: falling  # [falling, rising, both]
            topics-pub:  # optional
                button_pressed: /test/button/pressed
                button_state:   /test/button/state
            mqtt-translations:  # optional
                button_pressed: PRESSED
                button_state-open: OPEN
                button_state-closed: CLOSED
    """

    _pin = -1  # gpio pin id
    _sensor_type = None  # sensor type (one of the values in dict _sensor_type_list)
    _sensor_type_list = {  # list of valid sensor types. the sensor-type entry in yaml must be one of the keys.
        "DHT11": Adafruit_DHT.DHT11,
        "DHT22": Adafruit_DHT.DHT22,
        "AM2302": Adafruit_DHT.AM2302,
    }
    _calibrated_t = None  # copreus.baseclasses.CalibratedValue for temperature
    _calibrated_h = None  # copreus.baseclasses.CalibratedValue for humidity
    _event_pin = None  # input pin for poll_now trigger
    _event_flank_rising = None  # trigger poll_now on rising flank
    _event_flank_falling = None  # trigger poll_now on falling flank

    def __init__(self, config, mqtt_client=None, logger=None):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=__name__)
        APolling.__init__(self, self._config, self._mqtt_client, self._logger)
        AEvents.__init__(self, self._config, self._logger)

        self._pin = self._config["pin"]

        if self._config["sensor-type"] not in self._sensor_type_list.keys():
            self._logger.error("Wrong parameter. Value for 'sensor-type': {} is not in list of accepted values {}.".
                               format(self._config["sensor-type"], self._sensor_type_list.keys()))
            raise ValueError("Wrong parameter. Value for 'sensor-type': {} is not in list of accepted values {}.".
                             format(self._config["sensor-type"], self._sensor_type_list.keys()))
        self._sensor_type = self._sensor_type_list[self._config["sensor-type"]]

        self._calibrated_t = CalibratedValue(self._logger, self._config["calibration-temperature"], 1)
        self._calibrated_h = CalibratedValue(self._logger, self._config["calibration-humidity"], 1)

        try:
            self._event_pin = self._config["event-pin"]["pin"]
            if self._config["event-pin"]["flank"] == "rising":
                self._event_flank_rising = True
                self._event_flank_falling = False
            elif self._config["event-pin"]["flank"] == "falling":
                self._event_flank_rising = False
                self._event_flank_falling = True
            elif self._config["event-pin"]["flank"] == "both":
                self._event_flank_rising = True
                self._event_flank_falling = True
            else:
                self._logger.error("DHT.__init__ - unkown value for 'event-pin.flank' entry: '{}'.".
                                   format(self._config["event-pin"]["flank"]))
                raise ValueError("DHT.__init__ - unkown value for 'event-pin.flank' entry: '{}'.".
                                 format(self._config["event-pin"]["flank"]))
            # assign both dicts for a variable before calling init methods - this ensures that either both are present
            # or both are missing
            topics_pub = self._config["event-pin"]["topics-pub"]
            mqtt_translations = self._config["event-pin"]["mqtt-translations"]
            self._add_topics_pub(topics_pub)
            self._add_mqtt_translations(mqtt_translations)
            self._add_event(self._event_pin, self._callback_pin)
        except KeyError:
            pass

    def _callback_pin(self, channel):
        """The event pins state has changed - should poll_now be executed"""
        state = GPIO.input(self._event_pin)
        self._logger.info("DHT._callback_pin - received event. pin state: {}.".format(state))
        if not state:
            if self._event_flank_rising:
                self.poll_now()
            self._publish_value(self._topics_pub["button_pressed"], self._mqtt_translations["button_pressed"])
            self._publish_value(self._topics_pub["button_state"], self._mqtt_translations["button_state-closed"])
        else:
            if self._event_flank_falling:
                self.poll_now()
            self._publish_value(self._topics_pub["button_state"], self._mqtt_translations["button_state-open"])

    def _poll_device(self):
        """APolling._poll_device"""
        humidity, temperature = None, None
        while humidity is None or temperature is None:
            humidity, temperature = Adafruit_DHT.read_retry(self._sensor_type, self._pin)

        t = self._calibrated_t.value(temperature)
        h = self._calibrated_t.value(humidity)

        self._publish_value(self._topics_pub["temperature"], t)
        self._publish_value(self._topics_pub["humidity"], h)

    def _driver_start(self):
        """ADriver._driver_start"""
        self._start_polling()
        if self._event_pin:
            GPIO.setmode(GPIO.BCM)
            GPIO.setup(self._event_pin, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            self._register_events()

    def _driver_stop(self):
        """ADriver._driver_stop"""
        self._stop_polling()
        if self._event_pin:
            self._unregister_events()
            GPIO.cleanup(self._event_pin)

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method DHT.standalone()."""
    DHT.standalone()


if __name__ == "__main__":
    DHT.standalone()


import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from time import sleep
import _thread

from PIL import Image
from PIL import ImageDraw

from copreus.baseclasses.aepaper import AEPaper, EPaperConstants, EPaperMQTTMessageConverter
from tests_manual.tools import get_mqtt, get_model

print("send a lot of white images to the epaper - hopefully removing all 'ghosts'.")

filename = 'config_epaperdirect.yaml'
number_of_loops = 4
model = get_model(filename)

_ep_thread = _thread.start_new_thread(AEPaper.standalone, (['-c', filename],))

mqtt = get_mqtt(filename)
mqtt.connect()

white_image = Image.new('1', (EPaperConstants.model[model]["height"], EPaperConstants.model[model]["width"]), 255)  # 255: clear the frame
black_image = Image.new('1', (EPaperConstants.model[model]["height"], EPaperConstants.model[model]["width"]), 0)  # 0: clear the frame
white_draw = ImageDraw.Draw(white_image)
black_draw = ImageDraw.Draw(black_image)
msg_white = EPaperMQTTMessageConverter.to_full_image(white_image)
msg_black = EPaperMQTTMessageConverter.to_full_image(black_image)

print("sending {}x a white-black full image combinations to the epaper.".format(number_of_loops))
for i in range(0,number_of_loops):
    print("send black image")
    mqtt.client.publish("/w/thermostat/display/full_image_twice", msg_black)
    sleep(5)
    print("send white image")
    mqtt.client.publish("/w/thermostat/display/full_image_twice", msg_white)
    sleep(5)

mqtt.disconnect()
print("done")



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from copreus.drivers.pollinginput import PollingInput

from pelops.myconfigtools import read_config
from pelops.mymqttclient import MyMQTTClient
from pelops.mylogger import create_logger


def input_handler_pressed(message):
    print("button_pressed: {}".format(message))


def input_handler_state(message):
    print("button_state: {}".format(message))


filename = '../tests_unit/config_pollinginput.yaml'
config = read_config(filename)
logger = create_logger(config["logger"], __name__)
mqttclient = MyMQTTClient(config["mqtt"], logger, True)
mqttclient.connect()
mqttclient.subscribe(config["device"]["topics-pub"]["button_pressed"], input_handler_pressed)
mqttclient.subscribe(config["device"]["topics-pub"]["button_state"], input_handler_state)

PollingInput.standalone(('-c', filename))




import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["device"]["required"].extend(schema_part["required"])
    schema["device"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = {
        "pin": {
            "description": "gpio @ raspberry.",
            "type": "integer"
        },
        "sensor-type": {
            "description": "DHT11, DHT22, AM2302",
            "type": "string",
            "enum": ["DHT11", "DHT22", "AM2302"]
        },
        "event-pin": {
            "description": "input pin trigger for poll_now (optional)",
            "type": "object",
            "properties": {
                "pin": {
                    "description": "gpio pin",
                    "type": "integer"
                },
                "flank": {
                    "description": "trigger poll_now on which flank of the pin signal",
                    "type": "string",
                    "enum": ["rising", "falling", "both"]
                },
                "topics-pub": {
                    "description": "list of topics that the driver publishes to.",
                    "type": "object",
                    "properties": {
                        "button_pressed": {
                            "description": "mqtt-translations.button_pressed",
                            "type": "string"
                        },
                        "button_state": {
                            "description": "mqtt-translations.button_state-open and mqtt-translations.button_state-closed",
                            "type": "string"
                        }
                    },
                    "required": [
                        "button_pressed",
                        "button_state"
                    ],
                    "additionalItems": False
                },
                "mqtt-translations": {
                    "description": "what are the commands/states-values that are transmitted via mqtt",
                    "type": "object",
                    "properties": {
                        "button_pressed": {
                            "description": "value for button pressed",
                            "type": "string"
                        },
                        "button_state-open": {
                            "description": "value for button state",
                            "type": "string"
                        },
                        "button_state-closed": {
                            "description": "value for button state",
                            "type": "string"
                        }
                    },
                    "required": [
                        "button_pressed",
                        "button_state-open",
                        "button_state-closed"
                    ],
                    "additionalItems": False
                },
            },
            "required": ["pin", "flank"],
            "additionalProperties": False
        }
    }

    topics_pub = {
        "temperature": "raw temperature",
        "humidity": "raw humidity"
    }

    apolling_schema_parts, topics_sub, mqtt_translations = apolling.get_schema_parts()
    schema = adriver.get_schema("dht", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)
    schema["device"]["required"].remove("event-pin")
    _add_schema_part(schema, apolling_schema_parts)
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-humidity"))
    _add_schema_part(schema, calibratedvalue.get_schema_parts("calibration-temperature"))

    return schema



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import _thread
from time import sleep

from copreus.drivers.output import Output
from pelops.myconfigtools import read_config
from pelops.mymqttclient import MyMQTTClient
from pelops.mylogger import create_logger

filename = '../tests_unit/config_output.yaml'
config = read_config(filename)
logger = create_logger(config["logger"], __name__)
mqttclient = MyMQTTClient(config["mqtt"], logger, True)
mqttclient.connect()

def loop(topics, cmds):
    sleep(2)
    topic = topics["closed"]
    cmd_true = cmds["closed-true"]
    cmd_false = cmds["closed-false"]
    state = 0
    while 1:
        state = (state + 1) % 2
        if state:
            cmd = cmd_false
        else:
            cmd = cmd_true
        print(cmd)
        mqttclient.publish(topic, cmd)
        sleep(3)


_loop_thread = _thread.start_new_thread(loop, (config["device"]["topics-sub"], config["device"]["mqtt-translations"],))
Output.standalone(('-c', filename))




from copreus.baseclasses.adriver import ADriver
from copreus.baseclasses.aspi import ASPI
from copreus.baseclasses.calibratedvalue import CalibratedValue
from copreus.schema.dac import get_schema

class DAC(ADriver, ASPI):
    """Driver for ADC (analog-digital-converter) that are connected via spi (e.g. MCP4922).

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_sub: raw (=integer value for the dac), volt (=converted and calibrated value)
      * ASPI entries
      * CalibratedValue entries in a sub-block named 'calibration'
      * ADC own entries are
        * maxvalue: maximum value in volt. volt will be normalized towards this value.
        * bit: how many bits are used. typical values are 8, 10, and 12.
        * config_dac: configuration bit-sequence according to datasheet (0 if none)

    Example:
    device:
        type: dac
        config_dac: 0x3000
        spi:
            pin_cs: 4
            bus: 0
            device: 1
            maxspeed: 500000
        topics-sub:
            raw: /dac/raw
            volt: /dac/volt
        maxvalue: 24
        bit: 12
        calibration:
            use-calibration: True
            values:
            # - [ref_value, raw_value]
              - [0.0, 0.0]
              - [7.23, 6.0]
              - [24, 24]
    """

    _max_value = -1  # maximum value in volt
    _config_dac = -1  # configuration bit sequence according to data sheet
    _resolution = -1  # 2**bit
    _calibrated_value = None  # copreus.baseclasses.CalibratedValue

    def __init__(self, config, mqtt_client=None, logger=None, spi_lock=None):
        ADriver.__init__(self, config, mqtt_client, logger, logger_name=__name__)
        ASPI.__init__(self, self._config, self._logger, spi_lock)

        self._config_dac = self._config["config_dac"]

        self._max_value = self._config["maxvalue"]
        self._resolution = 2**int(self._config["bit"])
        self._calibrated_value = CalibratedValue(self._logger, self._config["calibration"], self._max_value / float(self._resolution))

        self._mqtt_client.subscribe(self._topics_sub["raw"], self._write_raw)
        self._mqtt_client.subscribe(self._topics_sub["volt"], self._write_volt)

    def _write_raw(self, msg):
        """on_message handler for topic sub 'raw'."""
        self._logger.info("received message '{}' on topic '{}'.".format(msg, self._topics_sub["raw"]))
        self._transfer_raw(int(msg))

    def _write_volt(self, msg):
        """on_message handler for topic sub 'volt'."""
        self._logger.info("received message '{}' on topic '{}'.".format(msg, self._topics_sub["volt"]))
        volt = float(msg)
        raw = self._calibrated_value.raw(volt)
        self._transfer_raw(raw)

    def _transfer_raw(self, raw):
        """Transfer the raw integer value (0<=value<2**bit) to the dac."""
        output = self._config_dac
        if raw > 4095:
            raw = 4095
        if raw < 0:
            raw = 0
        output |= raw
        buf0 = (output >> 8) & 0xff
        buf1 = output & 0xff
        self._logger.info("sending [{},{}] to dac.".format(buf0, buf1))
        self._transfer([buf0, buf1])

    def _driver_start(self):
        """ADriver._driver_start"""
        self._connect_spi()

    def _driver_stop(self):
        """ADriver._driver_stop"""
        self._disconnect_spi()

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method DAC.standalone()."""
    DAC.standalone()


if __name__ == "__main__":
    DAC.standalone()



from threading import Lock
import RPi.GPIO as GPIO
import spidev


class ASPI:
    """Additional base class for driver that want to use the SPI interface.

    Expects a sub-segment called 'spi' in the device yaml. The entries in this sub-segment are:
      * pin_cs - cable select gpio pin (if -1 or not pressent than spi-driver internal cs will be used. e.g.
      SPI_CE0_N/GPIO08 for device 0.)
      * bus - spi bus
      * device - spi device
      * maxspeed - maximum connection speed in Hz

    For the case that different devices are connected to the same bus/device but different cs-pins, this class uses a
    Lock mechanism to avoid concurrent resource usage. If none is provided to the constructor, each driver instance
    is acting independently.

    GPIO.setmode(GPIO.BCM)"""

    _cs = -1  # cable select gpio pin id
    _use_spi_internal_cs = False
    _bus = -1  # spi bus
    _device = -1  # spi device
    _max_speed = -1  # maximum connection speed in Hz
    _spi = None  # instance of spidev
    _spi_lock = None  # threading.Lock
    _logger_aspi = False  # print debugging information if set to yes.

    def __init__(self, config, logger, spi_lock = None):
        self._logger_aspi = logger
        if spi_lock is None:
            self._spi_lock = Lock()
        else:
            self._spi_lock = spi_lock

        try:
            self._use_spi_internal_cs = False
            self._cs = config["spi"]["pin_cs"]
            if self._cs == -1:
                self._use_spi_internal_cs = True
        except KeyError:
            self._use_spi_internal_cs = True

        self._bus = config["spi"]["bus"]
        self._device = config["spi"]["device"]
        self._max_speed = config["spi"]["maxspeed"]

        GPIO.setmode(GPIO.BCM)
        if not self._use_spi_internal_cs:
            GPIO.setup(self._cs, GPIO.OUT)

    def _connect_spi(self):
        """Initializes and opens spi interface."""
        self._logger_aspi.info("connect to spi")
        self._spi = spidev.SpiDev()
        self._spi.open(self._bus, self._device)
        self._spi.max_speed_hz = self._max_speed
        self._spi.mode = 0b00
        # self._spi.no_cs = True  # TODO - https://gitlab.com/tgd1975/copreus/issues/1

    def _disconnect_spi(self):
        """Closes spi interface."""
        self._logger_aspi.info("ASPI._disconnect_spi - disconnect from spi")
        self._spi.close()

    def _transfer(self, msg):
        """Start transfer if and only if the lock was aqcuired successfully. CS pin is lowered and raised before and
        after the transfer."""
        self._logger_aspi.debug("ASPI._transfer - acquire _spi_lock")
        with self._spi_lock:
            self._logger_aspi.debug("ASPI._transfer - ... acquire _spi_lock success")
            if not self._use_spi_internal_cs:
                GPIO.output(self._cs, 0)
            result = self._spi.xfer2(msg)
            if not self._use_spi_internal_cs:
                GPIO.output(self._cs, 1)
        self._logger_aspi.debug("ASPI._transfer - released lock _spi_lock")
        return result

    def _start_spi(self):
        """Initialize CS-pin and connect to spi interface. Usually called in the _start_sequence method of the
                silbling."""
        if not self._use_spi_internal_cs:
            GPIO.output(self._cs, 1)
        self._connect_spi()

    def _stop_spi(self):
        """Relase CS-pin and disconnect to spi interface. Usually called in the _stop_sequence method of the
                silbling."""
        self._disconnect_spi()
        if not self._use_spi_internal_cs:
            GPIO.cleanup(self._cs)



import copreus.schema.adc
import copreus.schema.bme_280
import copreus.schema.dac
import copreus.schema.dht
import copreus.schema.epaperdirect
import copreus.schema.epapersimple
import copreus.schema.input
import copreus.schema.pollinginput
import copreus.schema.output
import copreus.schema.rotaryencoder
import copreus.schema.rotaryencoder2


def _add_device_schema(schema, device_schema):
    device_schema["device"]["properties"]["active"] = {
        "description": "if set to false, the driver will not be loaded",
        "type": "boolean"
    }
    device_schema["device"]["required"].append("name")
    device_schema["device"]["required"].append("active")
    schema["devices"]["items"]["oneOf"].append(device_schema["device"])


def get_schema():
    schema = {
                "devices": {
                    "description": "Devicemanager configuration.",
                    "type": "array",
                    "items": {
                        "oneOf": [
                        ]
                    },
                    "additionalItems": False
                }
            }

    _add_device_schema(schema, copreus.schema.adc.get_schema())
    _add_device_schema(schema, copreus.schema.bme_280.get_schema())
    _add_device_schema(schema, copreus.schema.dac.get_schema())
    _add_device_schema(schema, copreus.schema.dht.get_schema())
    _add_device_schema(schema, copreus.schema.epaperdirect.get_schema())
    _add_device_schema(schema, copreus.schema.epapersimple.get_schema())
    _add_device_schema(schema, copreus.schema.input.get_schema())
    _add_device_schema(schema, copreus.schema.pollinginput.get_schema())
    _add_device_schema(schema, copreus.schema.output.get_schema())
    _add_device_schema(schema, copreus.schema.rotaryencoder.get_schema())
    _add_device_schema(schema, copreus.schema.rotaryencoder2.get_schema())

    return schema




from copreus.baseclasses.aepaper import AEPaper
from copreus.baseclasses.aepaper import EPaperMQTTMessageConverter
from time import sleep
import queue
import threading
from enum import Enum
from copreus.schema.epaperdirect import get_schema


class TaskType(Enum):
    """Task type for entries in the message queue."""
    DISPLAY = 0
    SWITCH = 1


class EPaperDirect(AEPaper):
    """
    This driver basically updates either the whole display (full_image) or selected parts (partial_image). The latter
    one results in faster update times. Internally, the epaper has two buffers. Via a command message the buffers are
    flipped. After flipping, one buffer is used to update the display while the other buffer is ready to receive new
    data via spi.

    When using partial image updates please take the two buffers under consideration. If two different areas are
    updated alternatively, it will result in a "blinking" behavior. The most common case - one static background
    image - and constantly update of the same area can be realised by first sending the full_image_twice and then
    the partial_image messages.

    Partial images must have a width and an x-position value that are multiples of eight. Any other value will result
    in a ValueError. Some displays have a width that is not compliant to this rule. In this case the display will have
    a logic width (e.g. 2.13 inch display has a width of 122 and a logic width of 128).

    The driver entry in the yaml file consists of:
      * ADriver entries
        * topics_sub:
            * full_image - a single image covering the whole display to be placed in the current buffer.
            * partial_image - list of image covering only parts of the display plus their position to be placed into
            the current buffer.
            * switch_frame - switch between the two frame buffers
        * mqtt-translations:
            * switch_frame - the command expected for switch_frame action
        * topics_pub:
            * message_queue_size - publishes the number of messages that wait to be processes.

    Example:
        device:
            type: epaperdirect
            model: 2.9
            spi:
                pin_cs: -1 # use spi cs mechanism. GPIO08/SPI_CE0_N
                bus: 0
                device: 0
                maxspeed: 2000000
            transpose: 270
            pin_rst:  13
            pin_dc:   19
            pin_busy: 6
            VCOM: -3.2
            autodeepsleep: True
            topics-sub:
                full_image: /test/display/full_image
                partial_image: /test/display/partial_image
                switch_frame: /test/display/switch_frame
            mqtt-translations:
                switch_frame: SWITCH
            topics-pub:
                message_queue_size: /test/display/message_queue_size

    """
    _msg_queue_size = 0  # number of mqtt messages that wait to be processed
    _msg_queue = None  # queue with tasks to be executed that are received via mqtt
    _msg_queue_worker_thread = None  # thread that processes all entries that are put to _msg_queue

    _topic_pub_msg_queue_size = None  # topic the current message queue size will be published to
    _switch_command = None  # expected command that initiates a frame buffer switch

    _display_frame_lock = None   # locked during image update (processing time + update_time)

    def __init__(self, config, mqtt_client=None, logger=None, spi_lock=None):
        """
        Constructor

        :param config: yaml config structure
        :param mqtt_client: mymqttclient instance (optional)
        :param logger: logger instance (optional)
        :param spi_lock: spi lock instance (optional)
        """

        AEPaper.__init__(self, config, mqtt_client, logger, spi_lock)
        self._display_frame_lock = threading.Lock()

        self._msg_queue = queue.Queue()
        self._msg_queue_size = 0
        self._msg_queue_worker_thread = threading.Thread(target=self._msg_queue_worker)

        self._switch_command = self._mqtt_translations["switch_frame"]
        self._topic_pub_msg_queue_size = self._topics_pub["message_queue_size"]

    def _handler_switch_frame(self, msg):
        """on_message handler for topic sub 'switch_frame'"""
        if msg == self._switch_command:
            self._logger.info("EPaperDirect._handler_switch_frame - received switch frame command via topic {}.".format(self._topics_sub["switch_frame"]))
            self._put_to_msg_queue(TaskType.SWITCH)
        else:
            self._logger.info("EPaperDirect._handler_switch_frame - received unknown command '{}' via topic {}. expected command {}."
                              .format(msg, self._topics_sub["switch_frame"], self._switch_command))

    def _handler_display_full_image(self, msg):
        """on_message handler for topic sub 'full_image'"""
        self._logger.info("EPaperDirect._handler_display_full_image - received full_image in topic {}.".format(self._topics_sub["full_image"]))
        image_entry = EPaperMQTTMessageConverter.from_full_image(msg, self._transpose,
                                                                 self._width, self._height)
        self._put_to_msg_queue(TaskType.DISPLAY, image_entry)

    def _handler_display_partial_image(self, msg):
        """on_message handler for topic sub 'partial_image'"""
        self._logger.info("EPaperDirect._handler_display_partial_image - received partial_image in topic {}.".format(self._topics_sub["partial_image"]))
        image_entries = EPaperMQTTMessageConverter.from_partial_images(msg, self._transpose,
                                                                       self._width, self._height)
        self._put_to_msg_queue(TaskType.DISPLAY, image_entries)

    def _msg_queue_worker(self):
        """
        process each item in queue and decrease _queue_size (new value is published). two different types of tasks
        can be processed:
          * DISPLAY - take the provided value and call _display_image
          * SWITCH - call _switch_frame
        this approach ensures that an incoming switch statement is processed if and only if all previously received
        images (full or partial) have been processed.
        """
        while True:
            task = self._msg_queue.get()
            if task is None:
                break

            tasktype, value = task

            if tasktype == TaskType.DISPLAY:
                self._display_image(value)
            elif tasktype == TaskType.SWITCH:
                self._switch_frame()
            else:
                self._logger.error("EPaperDirect._msg_queue_worker - unknown task type '{}'".format(tasktype))
                raise ValueError("EPaperDirect._msg_queue_worker - unknown task type '{}'".format(tasktype))

            self._msg_queue.task_done()
            self._msg_queue_size = self._msg_queue_size - 1

            self._logger.info("EPaperDirect._msg_queue_worker - mqtt message queue size decreased to: {}.".format(self._msg_queue_size))
            self._publish_value(self._topic_pub_msg_queue_size, self._msg_queue_size)

    def _put_to_msg_queue(self, tasktype, value=None):
        """
        increase queue size, publish new value and put task togehter with task type to _msg_queue

        :param tasktype: TaskType enum value
        :param value: payload for task (optional)
        :return:
        """
        self._msg_queue.put([tasktype, value])
        self._msg_queue_size = self._msg_queue_size + 1
        self._logger.info("EPaperDirect._put_to_msg_queue - mqtt message queue size increased to: {}.".format(self._msg_queue_size))
        self._publish_value(self._topic_pub_msg_queue_size, self._msg_queue_size)

    def _empty_msg_queue(self):
        """remove all tasks from _msg_queue"""
        while not self._msg_queue.empty():
            try:
                temp = self._msg_queue.get_nowait()
                self._msg_queue.task_done()
            except queue.Empty:
                pass
        self._msg_queue_size = 0
        self._logger.info("EPaperDirect._empty_msg_queue - mqtt message queue cleared")
        self._publish_value(self._topic_pub_msg_queue_size, self._msg_queue_size)

    def _switch_frame(self):
        """
        switch between the two frames. the active frame will be displayed.
        """
        self._logger.info("EPaperDirect._switch_frame - switch to other frame (_display_frame_lock)")
        with self._display_frame_lock:
            self._logger.info("EPaperDirect._switch_frame - _display_frame_lock acquired")
            self._display_frame()
            self._logger.info("EPaperDirect._switch_frame - _display_frame_lock released.")

    def _display_image(self, images):
        """send received image(s) to the epaper.

        acquire the lock (_display_frame_lock), wake up from deep sleep (optional), put all (partial) images into the current frame buffer,
        display frame, send back to deep sleep (optional), wait the _update_time and then release the lock."""
        self._logger.info("EPaperDirect._display_image - display image")
        with self._display_frame_lock:
            self._logger.info("EPaperDirect._display_image - lock _display_frame_lock acquired")
            if self._auto_deep_sleep:
                self._reset()

            for map in images:
                x = map["x"]
                y = map["y"]
                image = map["image"]
                self._logger.info("EPaperDirect._display_image - ... added image at {}/{}.".format(x, y))
                self._set_frame_memory(image, x, y)

            self._display_frame()
            if self._auto_deep_sleep:
                self._deep_sleep()
            sleep(self._update_time)
            self._logger.info("EPaperDirect._display_image - ... image displayed. lock _display_frame_lock released.")

    def _epaper_start(self):
        """start the message queue workers and subscribe to the incoming topics."""
        self._msg_queue_worker_thread.start()
        self._mqtt_client.subscribe(self._topics_sub["full_image"], self._handler_display_full_image)
        self._mqtt_client.subscribe(self._topics_sub["partial_image"], self._handler_display_partial_image)
        self._mqtt_client.subscribe(self._topics_sub["switch_frame"], self._handler_switch_frame)
        self._publish_value(self._topic_pub_msg_queue_size, self._msg_queue_size)

    def _epaper_stop(self):
        """unsubscribe from all topics and empty message queue"""
        with self._display_frame_lock:
            self._mqtt_client.unsubscribe(self._topics_sub["full_image"], self._handler_display_full_image)
            self._mqtt_client.unsubscribe(self._topics_sub["partial_image"], self._handler_display_partial_image)
            self._mqtt_client.unsubscribe(self._topics_sub["switch_frame"], self._handler_switch_frame)
            self._empty_msg_queue()
            self._msg_queue.join()

    @classmethod
    def _get_schema(cls):
        return get_schema()


def standalone():
    """Calls the static method EPaperDirect.standalone()."""
    EPaperDirect.standalone()


if __name__ == "__main__":
    EPaperDirect.standalone()



import copreus.schema.adriver as adriver
import copreus.schema.apolling as apolling
import copreus.schema.aspi as aspi
import copreus.schema.calibratedvalue as calibratedvalue


def _add_schema_part(schema, schema_part):
    schema["device"]["required"].extend(schema_part["required"])
    schema["device"]["properties"].update(schema_part["properties"])


def get_schema():
    driver_specific_properties = {
        "pin_sw": {
            "description": "gpio pin",
            "type": "integer"
        },
        "pin_dt": {
            "description": "gpio pin",
            "type": "integer"
        },
        "pin_clk": {
            "description": "gpio pin",
            "type": "integer"
        },
    }

    topics_pub = {
        "rotate": "mqtt-translations.rotate-left and mqtt-translations.rotate-right",
        "button_pressed": "mqtt-translations.button_pressed",
        "button_state": "mqtt-translations.button_state-open and mqtt-translations.button_state-closed"
    }
    topics_sub = {
    }
    mqtt_translations = {
        "rotate-left": "value for rotate left",
        "rotate-right": "value for rotate right",
        "button_pressed": "value for button pressed",
        "button_state-open": "value for button state",
        "button_state-closed": "value for button state",
    }

    schema = adriver.get_schema("rotaryencoder", driver_specific_properties, topics_pub, topics_sub, mqtt_translations)

    return schema
